from scNodes.core import config as cfg
import imgui
import time
import datetime
from scNodes.core.dataset import *
from scNodes.core import settings


class Node:
    title = "NullNode"
    group = "Ungrouped"
    colour = (1.0, 1.0, 1.0, 1.0)
    id_generator = count(0)
    size = 200
    sortid = 9999
    enabled = True

    COLOUR_WINDOW_BACKGROUND = (0.96, 0.96, 0.96, 0.96)
    COLOUR_FOCUSED_NODE_WINDOW_BACKGROUND = (0.99, 0.93, 0.93, 0.96)
    COLOUR_WINDOW_BORDER = (0.45, 0.45, 0.45, 1.0)
    COLOUR_WINDOW_BORDER_ACTIVE_NODE = (0.25, 0.25, 0.25, 1.0)
    COLOUR_WINDOW_BORDER_FOCUSED_NODE = (0 / 255, 0 / 255, 0 / 255)
    ACTIVE_NODE_BORDER_THICKNESS = 1.3
    FOCUSED_NODE_BORDER_THICKNESS = 1.1
    DEFAULT_NODE_BORDER_THICKNESS = 1.0
    COLOUR_TEXT = (0.0, 0.0, 0.0, 1.0)
    COLOUR_TEXT_DISABLED = (0.2, 0.4, 0.2, 1.0)
    COLOUR_FRAME_BACKGROUND = (0.84, 0.84, 0.84, 1.0)

    WINDOW_ROUNDING = 5.0
    FRAME_ROUNDING = 2.0
    PLAY_BUTTON_SIZE = 40
    PLAY_BUTTON_ICON_SIZE = 5.0
    PLAY_BUTTON_ICON_LINEWIDTH = 12.0
    PLAY_BUTTON_ICON_COLOUR = (0.2, 0.8, 0.2, 1.0)
    PLAY_BUTTON_WINDOW_BACKGROUND = (0.96, 0.96, 0.96, 1.0)
    PROGRESS_BAR_HEIGHT = 10
    PROGRESS_BAR_PADDING = 8

    TOOLTIP_APPEAR_DELAY = 1.0  # seconds
    TOOLTIP_HOVERED_TIMER = 0.0
    TOOLTIP_HOVERED_START_TIME = 0.0

    DISABLE_FRAME_INFO_WINDOW = False

    def __init__(self):
        self.id = int(datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')+"000") + next(Node.id_generator)
        self.position = [0, 0]
        self.last_measured_window_position = [0, 0]
        self.node_height = 100
        self.connectable_attributes = list()
        self.play = False
        self.any_change = False
        self.queued_actions = list()
        self.use_roi = False
        self.roi = [0, 0, 0, 0]
        self.lut = "auto"
        cfg.nodes.append(self)

        # some bookkeeping vars - ignore these
        self.buffer_last_output = False
        self.last_index_requested = -1
        self.last_frame_returned = None

        self.keep_active = False
        self.does_profiling_time = True
        self.does_profiling_count = True
        self.profiler_time = 0.0
        self.profiler_count = 0

        # flags
        self.NODE_RETURNS_IMAGE = True
        self.NODE_IS_DATA_SOURCE = False
        self.ROI_MUST_BE_SQUARE = False
        self.NODE_GAINED_FOCUS = False
        self.FRAME_REQUESTED_BY_IMAGE_VIEWER = False  # bit of a misc one. when the image viewer is requesting a frame (rather than the processing pipeline asking for one), this flag is temporarily set to True. This is used in e.g. the CropImage node, where the output to the image viewer is different than the output to the next node.
        self.OVERRIDE_AUTOCONTRAST = False
        self.OVERRIDE_AUTOCONTRAST_LIMS = (0, 65535)


    def __eq__(self, other):
        if isinstance(other, Node):
            return self.id == other.id
        return False

    def clear_flags(self):
        for a in self.connectable_attributes:
            a.clear_flags()
        if self.any_change and not cfg.focused_node == self:
            cfg.any_change = True
        self.any_change = False
        self.NODE_GAINED_FOCUS = False

    def render_start(self):
        if self.use_roi == False:
            self.roi = [0, 0, 0, 0]
        ## render the node window
        imgui.set_next_window_position(self.position[0], self.position[1], imgui.ALWAYS)
        imgui.set_next_window_size_constraints((self.size, 50), (self.size, 1000))
        imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM, *self.colour)
        imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM_HOVERED, *self.colour)
        imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND, *self.colour)
        imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_ACTIVE, *self.colour)
        imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_COLLAPSED, *self.colour)
        imgui.push_style_color(imgui.COLOR_CHECK_MARK, *self.colour)
        imgui.push_style_color(imgui.COLOR_SLIDER_GRAB, *self.colour)
        imgui.push_style_color(imgui.COLOR_SLIDER_GRAB_ACTIVE, *self.colour_brighten(self.colour))
        if cfg.focused_node == self:
            imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *self.colour_whiten(self.colour))
            imgui.push_style_var(imgui.STYLE_WINDOW_BORDERSIZE, Node.FOCUSED_NODE_BORDER_THICKNESS)
            imgui.push_style_color(imgui.COLOR_BORDER, *Node.COLOUR_WINDOW_BORDER_FOCUSED_NODE)
            imgui.push_style_color(imgui.COLOR_BORDER_SHADOW, *Node.COLOUR_WINDOW_BORDER_FOCUSED_NODE)
        elif cfg.active_node == self:
            imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *Node.COLOUR_WINDOW_BACKGROUND)
            imgui.push_style_var(imgui.STYLE_WINDOW_BORDERSIZE, Node.ACTIVE_NODE_BORDER_THICKNESS)
            imgui.push_style_color(imgui.COLOR_BORDER, *Node.COLOUR_WINDOW_BORDER_ACTIVE_NODE)
            imgui.push_style_color(imgui.COLOR_BORDER_SHADOW, *Node.COLOUR_WINDOW_BORDER_ACTIVE_NODE)
        else:
            imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *Node.COLOUR_WINDOW_BACKGROUND)
            imgui.push_style_var(imgui.STYLE_WINDOW_BORDERSIZE, Node.DEFAULT_NODE_BORDER_THICKNESS)
            imgui.push_style_color(imgui.COLOR_BORDER, *Node.COLOUR_WINDOW_BORDER)
            imgui.push_style_color(imgui.COLOR_BORDER_SHADOW, *Node.COLOUR_WINDOW_BORDER)
        imgui.push_style_color(imgui.COLOR_TEXT, *Node.COLOUR_TEXT)
        imgui.push_style_color(imgui.COLOR_HEADER, *self.colour_whiten(self.colour))
        imgui.push_style_color(imgui.COLOR_HEADER_HOVERED, *self.colour_whiten(self.colour))
        imgui.push_style_color(imgui.COLOR_HEADER_ACTIVE, *self.colour_whiten(self.colour))
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND_HOVERED, *self.colour_whiten(self.colour))
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND_ACTIVE, *Node.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_BUTTON, *self.colour)
        imgui.push_style_color(imgui.COLOR_BUTTON_HOVERED, *self.colour_brighten(self.colour))
        imgui.push_style_color(imgui.COLOR_BUTTON_ACTIVE, *self.colour_whiten(self.colour))
        imgui.push_style_color(imgui.COLOR_TAB_HOVERED, *self.colour_brighten(self.colour))
        imgui.push_style_var(imgui.STYLE_WINDOW_ROUNDING, Node.WINDOW_ROUNDING)
        imgui.push_style_var(imgui.STYLE_GRAB_ROUNDING, Node.WINDOW_ROUNDING)
        imgui.push_style_var(imgui.STYLE_FRAME_ROUNDING, Node.FRAME_ROUNDING)
        imgui.push_style_var(imgui.STYLE_WINDOW_MIN_SIZE, (1, 1))
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND, *Node.COLOUR_FRAME_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_POPUP_BACKGROUND, *Node.COLOUR_WINDOW_BACKGROUND)
        _, stay_open = imgui.begin(self.title + f"##{self.id}", True, imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_BRING_TO_FRONT_ON_FOCUS | imgui.WINDOW_NO_SAVED_SETTINGS | imgui.WINDOW_NO_MOVE | imgui.WINDOW_ALWAYS_AUTO_RESIZE | imgui.WINDOW_NO_COLLAPSE)

        if imgui.is_window_focused() and not imgui.is_any_item_hovered():
            if cfg.set_active_node(self):
                self.on_gain_focus()
            self.position[0] = self.position[0] + cfg.node_move_requested[0]
            self.position[1] = self.position[1] + cfg.node_move_requested[1]
        self.position[0] += cfg.camera_move_requested[0]
        self.position[1] += cfg.camera_move_requested[1]
        self.last_measured_window_position = imgui.get_window_position()
        self.node_height = imgui.get_window_size()[1]
        imgui.push_clip_rect(0, 0, cfg.window_width, cfg.window_height)
        imgui.push_id(str(self.id))
        if not stay_open:
            self.render_end()
            self.delete()
            return False

        return True

    def render(self):
        if self.render_start():
            self.render_end()

    def render_end(self):
        ## Node context_menu:
        if imgui.begin_popup_context_window():
            self.keep_active = cfg.focused_node == self
            _changed, self.keep_active = imgui.checkbox("Focus node", self.keep_active)
            if _changed:
                if self.keep_active:
                    cfg.set_active_node(self, True)
                else:
                    cfg.set_active_node(None, True)
                    cfg.set_active_node(self, False)
                imgui.close_current_popup()
            # _duplicate, _ = imgui.menu_item("Duplicate node")
            # if _duplicate:
            #     duplicate_node = Node.create_node_by_type(self.type) ## TODO FIX
            #     duplicate_node.position = [self.position[0] + 10, self.position[1] + 10]
            #     cfg.set_active_node(duplicate_node)
            _delete, _ = imgui.menu_item("Delete node")
            if _delete:
                self.delete()
            _reset, _ = imgui.menu_item("Reset node")
            # if _reset:
            #     new_node = Node.create_node_by_type(self.type) ## TODO FIX
            #     new_node.position = self.position
            #     self.delete()
            if imgui.begin_menu("Set node-specific LUT"):
                for lut in ["auto"] + settings.lut_names:
                    _lut, _ = imgui.menu_item(lut)
                    if _lut:
                        self.lut = lut
                imgui.end_menu()
            imgui.end_popup()
        if cfg.profiling and self.does_profiling_time:
            imgui.separator()
            imgui.text(f"Time processing: {self.profiler_time:.2f}")
            imgui.text(f"Frames requested: {self.profiler_count}")
        imgui.pop_style_color(23)
        imgui.pop_style_var(5)
        imgui.pop_id()
        imgui.pop_clip_rect()
        imgui.end()

    def play_button(self):
        """
        Render a 'play / pause' button for the node, with play/pause state toggled by a single left mouse button click.
        :return: tuple (bool clicked, bool play). bool play is True is current state is 'play', False is 'pause'
        """
        clicked = False
        ## Start/stop button
        imgui.push_id(f"Startstop{self.id}")
        window_position = (self.last_measured_window_position[0] + self.size - Node.PLAY_BUTTON_SIZE / 2, self.last_measured_window_position[1] + self.node_height / 2 - Node.PLAY_BUTTON_SIZE / 2)

        imgui.set_next_window_position(*window_position)
        imgui.set_next_window_size(Node.PLAY_BUTTON_SIZE, Node.PLAY_BUTTON_SIZE)
        imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *Node.PLAY_BUTTON_WINDOW_BACKGROUND)
        imgui.push_style_var(imgui.STYLE_WINDOW_ROUNDING, 25.0)
        imgui.begin(f"##Node{self.id}playbtn", False,
                    imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_COLLAPSE)
        if imgui.is_window_hovered() and imgui.is_mouse_clicked(0, False):
            self.play = not self.play
            clicked = True
            self.any_change = True
        imgui.pop_style_color(1)
        imgui.pop_style_var(1)

        ## Draw play / pause symbols
        draw_list = imgui.get_window_draw_list()
        if self.play:
            # Draw STOP button
            origin = [window_position[0] + Node.PLAY_BUTTON_SIZE / 2.0,
                      window_position[1] + Node.PLAY_BUTTON_SIZE / 2.0]
            p1 = (-1.0 * Node.PLAY_BUTTON_ICON_SIZE, 1.5 * Node.PLAY_BUTTON_ICON_SIZE)
            p2 = (-1.0 * Node.PLAY_BUTTON_ICON_SIZE, -1.5 * Node.PLAY_BUTTON_ICON_SIZE)
            draw_list.add_polyline([(p1[0] + origin[0], p1[1] + origin[1]),
                                    (p2[0] + origin[0], p2[1] + origin[1])], imgui.get_color_u32_rgba(*Node.PLAY_BUTTON_ICON_COLOUR), False, Node.PLAY_BUTTON_ICON_LINEWIDTH / 2.0)
            p1 = (1.0 * Node.PLAY_BUTTON_ICON_SIZE, 1.5 * Node.PLAY_BUTTON_ICON_SIZE)
            p2 = (1.0 * Node.PLAY_BUTTON_ICON_SIZE, -1.5 * Node.PLAY_BUTTON_ICON_SIZE)
            draw_list.add_polyline([(p1[0] + origin[0], p1[1] + origin[1]),
                                    (p2[0] + origin[0], p2[1] + origin[1])],
                                   imgui.get_color_u32_rgba(*Node.PLAY_BUTTON_ICON_COLOUR), False,
                                   Node.PLAY_BUTTON_ICON_LINEWIDTH / 2.0)
        else:
            # Draw PLAY button
            origin = [window_position[0] + Node.PLAY_BUTTON_SIZE / 2.0,
                      window_position[1] + Node.PLAY_BUTTON_SIZE / 2.0]
            p1 = (Node.PLAY_BUTTON_ICON_SIZE * -0.65, Node.PLAY_BUTTON_ICON_SIZE)
            p2 = (Node.PLAY_BUTTON_ICON_SIZE * 0.85, 0)
            p3 = (Node.PLAY_BUTTON_ICON_SIZE * -0.65, -Node.PLAY_BUTTON_ICON_SIZE)
            draw_list.add_polyline([(p1[0] + origin[0], p1[1] + origin[1]),
                                    (p2[0] + origin[0], p2[1] + origin[1]),
                                    (p3[0] + origin[0], p3[1] + origin[1])], imgui.get_color_u32_rgba(*Node.PLAY_BUTTON_ICON_COLOUR), True, Node.PLAY_BUTTON_ICON_LINEWIDTH)
        imgui.end()
        imgui.pop_id()

        return clicked, self.play

    def progress_bar(self, progress):
        """
        Renders a progress bar at the current imgui cursor position.
        :param progress:float 0.0 to 1.0.
        :return:
        """
        width = imgui.get_content_region_available_width()
        origin = imgui.get_window_position()
        y = imgui.get_cursor_screen_pos()[1]
        drawlist = imgui.get_window_draw_list()
        drawlist.add_rect_filled(Node.PROGRESS_BAR_PADDING + origin[0], y,
                                 Node.PROGRESS_BAR_PADDING + origin[0] + width, y + Node.PROGRESS_BAR_HEIGHT,
                                 imgui.get_color_u32_rgba(*Node.colour_whiten(self.colour)))
        drawlist.add_rect_filled(Node.PROGRESS_BAR_PADDING + origin[0], y,
                                 Node.PROGRESS_BAR_PADDING + origin[0] + width * min([1.0, progress]),
                                 y + Node.PROGRESS_BAR_HEIGHT, imgui.get_color_u32_rgba(*self.colour))

    def delete(self):
        try:
            if cfg.focused_node == self:
                cfg.set_active_node(None, True)
            if cfg.active_node == self:
                cfg.set_active_node(None)
            if self in cfg.nodes:
                cfg.nodes.remove(self)
            for attribute in self.connectable_attributes:
                attribute.delete()
        except Exception as e:
            cfg.set_error(e, "Problem deleting node\n"+str(e))

    @staticmethod
    def tooltip(text):
        if imgui.is_item_hovered():
            if Node.TOOLTIP_HOVERED_TIMER == 0.0:
                Node.TOOLTIP_HOVERED_START_TIME = time.time()
                Node.TOOLTIP_HOVERED_TIMER = 0.001  # add a fake 1 ms to get out of this if clause
            elif Node.TOOLTIP_HOVERED_TIMER > Node.TOOLTIP_APPEAR_DELAY:
                imgui.set_tooltip(text)
            else:
                Node.TOOLTIP_HOVERED_TIMER = time.time() - Node.TOOLTIP_HOVERED_START_TIME
        if not imgui.is_any_item_hovered():
            Node.TOOLTIP_HOVERED_TIMER = 0.0

    @staticmethod
    def colour_brighten(c):
        return (c[0] * 1.2, c[1] * 1.2, c[2] * 1.2, 1.0)

    @staticmethod
    def colour_whiten(c):
        return (c[0] * 0.3 + 0.7, c[1] * 0.3 + 0.7, c[2] * 0.3 + 0.7, 1.0)

    @staticmethod
    def get_source_load_data_node(node):
        """
        This function is _static_ in order to ensure proper usage; it relies on a node having a connectable_attribute with type INPUT that ultimately is linked to a LoadDataNode.
        :param node: the node for which you want to find the source LoadDataNode.
        :return: node: node with type TYPE_LOAD_DATA
        """
        def get_any_incoming_node(_node):
            for attribute in _node.connectable_attributes:
                if attribute.direction == ConnectableAttribute.INPUT:
                    return attribute.get_incoming_node()

        if node.NODE_IS_DATA_SOURCE:
            return node
        try:
            source = get_any_incoming_node(node)
            while not isinstance(source, NullNode):
                if source.NODE_IS_DATA_SOURCE:
                    return source
                source = get_any_incoming_node(source)
            return source
        except Exception as e:
            return NullNode()

    def get_image(self, idx):
        """
        :param idx:int, index of frame in dataset
        :return: Frame object
        """
        retval = None
        if cfg.profiling:
            start_time = time.time()
            self.profiler_count += 1
        try:
            if self.buffer_last_output:
                if idx is self.last_index_requested:
                    retval = copy.deepcopy(self.last_frame_returned)
                else:
                    self.last_frame_returned = self.get_image_impl(idx)
                    self.last_index_requested = idx
                    retval = copy.deepcopy(self.last_frame_returned)
            retval = self.get_image_impl(idx)
        except Exception as e:
            cfg.set_error(e, f"{self} error: "+str(e))
        if cfg.profiling:
            self.profiler_time += (time.time() - start_time)
        return retval

    def get_image_impl(self, idx):
        return None

    def get_particle_data(self):
        return self.get_particle_data_impl()

    def get_particle_data_impl(self):
        return ParticleData()

    def get_coordinates(self, idx):
        return None

    def on_update(self):
        return None

    def on_gain_focus(self):
        self.NODE_GAINED_FOCUS = True

    def pre_save(self):
        cfg.pickle_temp = dict()
        self.last_index_requested = -1
        self.last_frame_returned = None
        cfg.pickle_temp["profiler_time"] = self.profiler_time
        self.pre_save_impl()

    def pre_save_impl(self):
        pass

    def post_save(self):
        self.profiler_time = cfg.pickle_temp["profiler_time"]
        self.post_save_impl()
        cfg.pickle_temp = dict()

    def post_save_impl(self):
        pass


class NullNode(Node):
    def __init__(self):
        super().__init__()
        self.dataset = Dataset()
        self.dataset_in = None
        self.pixel_size = 100
        cfg.nodes.remove(self)

    def __bool__(self):
        return False


class ConnectableAttribute:
    id_generator = count(0)
    active_connector = None

    TYPE_DATASET = 1
    TYPE_IMAGE = 2
    TYPE_RECONSTRUCTION = 3
    TYPE_COLOUR = 4
    TYPE_MULTI = 5
    TYPE_COORDINATES = 6

    INPUT = True
    OUTPUT = False

    COLOUR = dict()
    COLOUR[TYPE_DATASET] = (54 / 255, 47 / 255, 192 / 255, 1.0)
    COLOUR[TYPE_IMAGE] = (255 / 255, 179 / 255, 35 / 255, 1.0)
    COLOUR[TYPE_RECONSTRUCTION] = (243 / 255, 9 / 255, 9 / 255, 1.0)
    COLOUR[TYPE_COLOUR] = (4 / 255, 4 / 255, 4 / 255, 1.0)
    COLOUR[TYPE_MULTI] = (230 / 255, 230 / 255, 240 / 255, 1.0)
    COLOUR[TYPE_COORDINATES] = (230 / 255, 174 / 255, 13 / 255, 1.0)
    COLOUR_BORDER = (0.0, 0.0, 0.0, 1.0)

    TITLE = dict()
    TITLE[TYPE_DATASET] = "Dataset"
    TITLE[TYPE_IMAGE] = "Image"
    TITLE[TYPE_RECONSTRUCTION] = "Reconstruction"
    TITLE[TYPE_COLOUR] = "Colour"
    TITLE[TYPE_MULTI] = "Dataset"
    TITLE[TYPE_COORDINATES] = "Coordinates"

    CONNECTOR_SPACING = 10
    CONNECTOR_HORIZONTAL_OFFSET = -8
    CONNECTOR_VERTICAL_OFFSET = 7
    CONNECTOR_RADIUS = 5
    CONNECTOR_WINDOW_PADDING = 9
    CONNECTOR_SEGMENTS = 16
    CONNECTION_LINE_COLOUR = (0.0, 0.0, 0.0, 1.0)
    CONNECTION_LINE_THICKNESS = 2

    def __init__(self, attributetype, direction, parent, allowed_partner_types=None):
        self.id = int(datetime.datetime.utcnow().timestamp()) + next(ConnectableAttribute.id_generator)
        self.type = attributetype
        self.title = ConnectableAttribute.TITLE[self.type]
        self.colour = ConnectableAttribute.COLOUR[self.type]
        self.direction = direction
        self.linked_attributes = list()
        self.draw_y = 0
        self.draw_x = 0
        self.connector_position = [0, 0]
        self.newly_connected = False
        self.any_change = False
        self.newly_disconnected = False
        self.parent = parent
        self.multi = self.type == ConnectableAttribute.TYPE_MULTI
        self.allowed_partner_types = [self.type]
        self.current_type = self.type
        if allowed_partner_types is not None:
            self.allowed_partner_types = allowed_partner_types

        parent.connectable_attributes.append(self)

    def __eq__(self, other):
        if type(self) is type(other):
            return self.id == other.id
        return False

    def __str__(self):
        return f"Attribute type {self.type} with id {self.id}"

    def clear_flags(self):
        self.any_change = False

    def render_start(self):
        imgui.push_id(f"Attribute{self.id}")

    def render_end(self, show_label=True):
        text_width = imgui.get_font_size() * len(self.title) / 2
        window_width = imgui.get_window_content_region_width()
        self.draw_x = imgui.get_cursor_screen_pos()[0] + (ConnectableAttribute.CONNECTOR_HORIZONTAL_OFFSET if self.direction else window_width - 1 - ConnectableAttribute.CONNECTOR_HORIZONTAL_OFFSET)
        if self.direction == ConnectableAttribute.INPUT:
            imgui.same_line(position = ConnectableAttribute.CONNECTOR_SPACING)
            self.draw_y = imgui.get_cursor_screen_pos()[1] + ConnectableAttribute.CONNECTOR_VERTICAL_OFFSET
        else:
            imgui.same_line(position=window_width - text_width)
            self.draw_y = imgui.get_cursor_screen_pos()[1] + ConnectableAttribute.CONNECTOR_VERTICAL_OFFSET

        if show_label:
            imgui.text(self.title)
        imgui.pop_id()

        ## connector drawing
        self.connector_position = (self.draw_x, self.draw_y)
        draw_list = imgui.get_window_draw_list()
        draw_list.add_circle_filled(self.connector_position[0], self.connector_position[1],
                                    ConnectableAttribute.CONNECTOR_RADIUS + 1, imgui.get_color_u32_rgba(0.0, 0.0, 0.0, 1.0),
                                    ConnectableAttribute.CONNECTOR_SEGMENTS)
        draw_list.add_circle_filled(self.connector_position[0], self.connector_position[1], ConnectableAttribute.CONNECTOR_RADIUS, imgui.get_color_u32_rgba(*self.colour), ConnectableAttribute.CONNECTOR_SEGMENTS)

        ## Connector user input
        self.connector_drag_drop_sources()

        if self.direction == ConnectableAttribute.OUTPUT:
            for partner in self.linked_attributes:
                draw_list = imgui.get_background_draw_list()
                draw_list.add_line(self.draw_x, self.draw_y, partner.draw_x, partner.draw_y, imgui.get_color_u32_rgba(*ConnectableAttribute.CONNECTION_LINE_COLOUR), ConnectableAttribute.CONNECTION_LINE_THICKNESS)
        if cfg.active_connector == self:
            if cfg.active_connector_hover_pos is not None:
                draw_list = imgui.get_background_draw_list()
                draw_list.add_line(self.draw_x, self.draw_y, *cfg.active_connector_hover_pos, imgui.get_color_u32_rgba(*ConnectableAttribute.CONNECTION_LINE_COLOUR), ConnectableAttribute.CONNECTION_LINE_THICKNESS)
        if self.any_change:
            self.parent.any_change = True

    def get_incoming_node(self):
        if self.direction == ConnectableAttribute.INPUT:
            if len(self.linked_attributes) == 1:
                return self.linked_attributes[0].parent
        return NullNode()

    def get_incoming_attribute_type(self):
        if self.direction == ConnectableAttribute.INPUT:
            if len(self.linked_attributes) == 1:
                return self.linked_attributes[0].type
        return None

    def connector_drag_drop_sources(self):
        ## 'render' the invisible node drag/drop sources]
        connector_window_size = (ConnectableAttribute.CONNECTOR_RADIUS + ConnectableAttribute.CONNECTOR_WINDOW_PADDING) * 2 + 1
        imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, 0.0, 0.0, 0.0, 0.0)
        imgui.push_style_color(imgui.COLOR_BORDER, 0.0, 0.0, 0.0, 0.0)
        imgui.push_style_color(imgui.COLOR_BORDER_SHADOW, 0.0, 0.0, 0.0, 0.0)
        imgui.push_style_color(imgui.COLOR_DRAG_DROP_TARGET, 0.0, 0.0, 0.0, 1.0)

        imgui.set_next_window_size(connector_window_size, connector_window_size)
        imgui.set_next_window_position(self.draw_x - connector_window_size // 2,
                                       self.draw_y - connector_window_size // 2)
        imgui.begin(f"Attribute{self.id}", False, imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_COLLAPSE | imgui.WINDOW_NO_SAVED_SETTINGS | imgui.WINDOW_NO_SCROLLBAR | imgui.WINDOW_NO_SAVED_SETTINGS)

        if imgui.is_window_hovered():
            if cfg.connector_delete_requested:
                cfg.connector_delete_requested = False
                self.disconnect_all()
                self.any_change = True

        ## Allow draw source if currently no node being edited
        if cfg.active_connector is self or cfg.active_connector is None:
            if imgui.begin_drag_drop_source(imgui.DRAG_DROP_SOURCE_NO_PREVIEW_TOOLTIP):
                cfg.active_connector = self
                cfg.active_connector_parent_node = self.parent
                imgui.set_drag_drop_payload('connector', b'1')  # Arbitrary payload value - manually keeping track of the payload elsewhere
                imgui.end_drag_drop_source()
        elif cfg.active_connector is not None:
            imgui.begin_child(f"Attribute{self.id}drop_target")
            imgui.end_child()
            if imgui.begin_drag_drop_target():
                if cfg.connector_released:
                    self.connect_attributes(cfg.active_connector)
                    cfg.connector_released = False
                imgui.end_drag_drop_target()
        imgui.end()
        imgui.pop_style_color(4)

    def connect_attributes(self, partner):
        io_match = self.direction != partner.direction
        type_match = self.type == partner.type or partner.type in self.allowed_partner_types or self.type in partner.allowed_partner_types
        novel_match = partner not in self.linked_attributes
        parent_match = self.parent != partner.parent
        if io_match and type_match and novel_match and parent_match:
            self.any_change = True
            self.parent.any_change = True
            partner.any_change = True
            partner.parent.any_change = True
            self.linked_attributes.append(partner)
            partner.linked_attributes.append(self)
            self.on_connect()
            partner.on_connect()
        cfg.active_connector_parent_node = None

    def disconnect_all(self):
        for partner in self.linked_attributes:
            partner.linked_attributes.remove(self)
            partner.notify_disconnect()
        self.notify_disconnect()
        self.linked_attributes = list()
        self.parent.any_change = True
        self.colour = ConnectableAttribute.COLOUR[self.type]
        self.title = ConnectableAttribute.TITLE[self.type]

    def is_connected(self):
        return len(self.linked_attributes) > 1

    def delete(self):
        self.disconnect_all()

    def on_connect(self):
        ## when on connect is called, a new ConnectableAttribute has just been added to the linked_partners list
        if self.direction == ConnectableAttribute.INPUT:
            for idx in range(len(self.linked_attributes) - 1):
                self.linked_attributes[idx].notify_disconnect()
                self.linked_attributes[idx].linked_attributes.remove(self)
                self.linked_attributes.remove(self.linked_attributes[idx])
        self.newly_connected = True
        self.parent.any_change = True
        if self.type == ConnectableAttribute.TYPE_MULTI:
            self.colour = self.linked_attributes[-1].colour
            self.current_type = self.linked_attributes[-1].type
            self.title = self.linked_attributes[-1].title

    def notify_disconnect(self):
        self.any_change = True
        self.newly_disconnected = True
        self.colour = ConnectableAttribute.COLOUR[self.type]
        self.title = ConnectableAttribute.TITLE[self.type]

    def check_connect_event(self):
        if self.newly_connected:
            self.newly_connected = False
            return True
        return False

    def check_disconnect_event(self):
        if self.newly_disconnected:
            self.newly_disconnected = False
            return True
        return False





from itertools import count
import numpy as np
import re
import glob
from PIL import Image
from skimage import transform
import pandas as pd
import copy
import tifffile

class Dataset:
    idgen = count(1)

    def __init__(self, path=None, pixel_size=100):
        """
        :param path: None or string, path to either i) a multi-page .tif file with ordering XYF, F the number of frames, or ii) a single-page .tif file, in which case a Dataset is generated comprising all the .tif files in that folder. If 'None', a Dataset() object is generated that has no frame data.
        :param pixel_size:
        """
        self.id = next(Dataset.idgen)
        self.path = path
        self.frames = list()
        self.n_frames = 0
        self.current_frame = 0
        self.pixel_size = pixel_size
        self.initialized = False
        self.reconstruction_roi = [0, 0, 0, 0]
        if self.path is not None:
            self.directory = self.path[:self.path.rfind("/") + 1]
            self.load_data()
            self.img_width, self.img_height = self.get_indexed_image(0).load().shape
            self.initialized = True
        else:
            self.directory = ""
            self.img_width, self.img_height = (0, 0)

    def load_data(self):
        self.frames = list()

        def numerical_sort(value):
            numbers = re.compile(r'(\d+)')
            parts = numbers.split(value)
            parts[1::2] = map(int, parts[1::2])
            return parts

        # stack
        img = Image.open(self.path)
        if img.n_frames > 1:
            for i in range(img.n_frames):
                self.n_frames += 1
                self.frames.append(Frame(self.path, i, framenr=i))

        # folder
        else:
            files = sorted(glob.glob(self.directory + "*.tif*"), key=numerical_sort)
            for file in files:
                self.n_frames += 1
                self.frames.append(Frame(file, 0, framenr=self.n_frames - 1))

    def get_indexed_image(self, index):
        if 0 <= index < self.n_frames:
            self.frames[index].pixel_size = self.pixel_size
            return self.frames[index]
        else:
            if index <= 0:
                self.frames[0].pixel_size = self.pixel_size
                return self.frames[0]
            else:
                self.frames[-1].pixel_size = self.pixel_size
                return self.frames[-1]

    def get_active_image(self):
        self.frames[self.current_frame].pixel_size = self.pixel_size
        return self.frames[self.current_frame]

    def filter_frames_by_title(self, positive_filter_string, negative_filter_string):
        """
        :param positive_filter_string: string, all frames not containing any of the semicolon separated expressions in this string are dicarded
        :param negative_filter_string: string, all frames containing any of the semicolon separated expresions in this string are discarded.
        Note that when positive and negative tags contradict, preference is given to the positive filter. I.e., the frame is not discarded.
        :return: Nothing, dataset object itself is affected
        """
        discard = [False] * len(self.frames)
        neg_tags = negative_filter_string.split(';')
        pos_tags = positive_filter_string.split(';')
        while '' in neg_tags:
            neg_tags.remove('')
        while '' in pos_tags:
            pos_tags.remove('')
        i = 0
        for frame in self.frames:
            for tag in neg_tags:
                if tag in frame.path:
                    discard[i] = True
            for tag in pos_tags:
                if tag in frame.path:
                    discard[i] = False
                else:
                    discard[i] = True
            i += 1
        for i in range(len(self.frames) - 1, -1, -1):
            if discard[i]:

                self.frames.pop(i)
        self.n_frames = len(self.frames)

    def append_frame(self, frame):
        """
        Add a frame to the end of the list of frames in the dataset.
        :param frame: Frame object (see below).
        """
        if not self.initialized:
            self.img_width, self.img_height = frame.load().shape
            self.initialized = True
        self.frames.append(frame)
        self.n_frames += 1

    def delete_by_index(self, idx):
        if 0 <= idx < len(self.frames):
            del self.frames[idx]
            self.n_frames = len(self.frames)

    def __str__(self):
        return "Dataset object with source at: "+self.path+f"\nn_frames = {self.n_frames}"


class Frame:
    id_gen = count(1)

    def __init__(self, path, index=None, framenr=0):
        self.id = next(Frame.id_gen)
        self.index = index
        self.framenr = framenr
        self.path = path
        self.data = None
        self.raw_data = None
        self.width = None
        self.height = None
        self.discard = False
        self.translation = [0.0, 0.0]
        self.maxima = list()
        self.particles = list()
        self.pixel_size = 1
        self.gpu_data_buffer = None
        self.gpu_params_buffer = None
        self.gpu_crop_xy_buffer = None
        self._ce_lut = 0
        self._ce_clims = [0, 1]

    def load(self):
        if self.data is not None:
            return self.data
        elif self.raw_data is not None:
            self.data = self.raw_data.copy()
            return self.data
        else:
            # 221202 Note: using tifffile instead of PIL now!
            if self.index is None:
                self.raw_data = tifffile.imread(self.path).astype(np.float)
            else:
                self.raw_data = tifffile.imread(self.path, key=self.index).astype(np.float)
            self.data = self.raw_data.copy()
            self.width = self.raw_data.shape[0]
            self.height = self.raw_data.shape[1]
            return self.data

    def clean(self):
        self.translation = [0.0, 0.0]
        self.discard = False
        self.data = None ## TODO: check whether app is faster and output still correct when this line is removed.
        self.maxima = list()

    def clone(self):
        return copy.deepcopy(self)

    def bake_transform(self, interpolation = 1, edges = 'constant', preserve_range=False):
        tmat = transform.AffineTransform(np.matrix([[1.0, 0.0, self.translation[0]], [0.0, 1.0, self.translation[1]], [0.0, 0.0, 1.0]]))
        self.data = transform.warp(self.data, tmat, order=interpolation, mode=edges, preserve_range=preserve_range)

    def load_roi(self, roi=None):
        if roi is None:
            return self.load()
        else:
            return self.load()[roi[1]:roi[3], roi[0]:roi[2]]

    def __str__(self):
        sstr = f"Frame at path: \n{self.path}\n" \
            + ("Discarded frame. " if self.discard else "Frame in use. ") \
            + f"Shift of ({self.translation[0]:.2f}, {self.translation[1]:.2f}) pixels detected."
        if self.maxima is not []:
            sstr += f"\n{len(self.maxima)} particles found."
        return sstr


class ParticleData:
    HISTOGRAM_BINS = 50

    def __init__(self, pixel_size=100):
        self.pixel_size = pixel_size
        self.particles = list()
        self.n_particles = 0
        self.parameter = dict()
        self.histogram_counts = dict()
        self.histogram_bins = dict()
        self.x_max = 0
        self.y_max = 0
        self.x_min = 0
        self.y_min = 0
        self.baked = False  # flag to store whether function bake() has been performed after any changes to the particle list.
        self.baked_by_renderer = False  # flag to store whether renderer has created instance buffers using the latest particle list. ONLY renderer can set it to True, any edits from within ParticleData set it to false.
        self.colours_up_to_date = True
        self.empty = True
        self.reconstruction_roi = [0, 0, 0, 0]

    def __add__(self, other):
        if isinstance(other, list):
            for p in other:
                self.particles.append(p)
            self.baked = False
            self.baked_by_renderer = False
            self.empty = False
            return self

    def __str__(self):
        r = self.reconstruction_roi
        return f"ParticleData obj. with {self.n_particles} particles.\nx range = ({self.x_min}, {self.x_max}), y range = ({self.y_min}, {self.y_max})\nCorresponding ROI in input source image: x {r[0]} to {r[2]}, y {r[1]} to {r[3]}."

    def clean(self):
        for particle in self.particles:
            particle.visible = True
        self.baked_by_renderer = False

    def bake(self, discard_filtered_particles=False):
        if self.empty:
            return
        self.baked = True
        self.baked_by_renderer = False
        frame = list()
        x = list()
        y = list()
        sigma = list()
        intensity = list()
        offset = list()
        uncertainty = list()
        bkgstd = list()

        k1 = self.pixel_size**2 / 12
        k2 = 8 * np.pi / self.pixel_size**2
        self.n_particles = 0
        for p in self.particles:
            if p.visible or not discard_filtered_particles:
                if p.intensity == 0.0:
                    continue
                self.n_particles += 1
                frame.append(p.frame)
                x.append(p.x)
                y.append(p.y)
                sigma.append(p.sigma * self.pixel_size)
                intensity.append(p.intensity)
                offset.append(p.offset)
                if p.uncertainty == -1:
                    p.uncertainty = np.sqrt(((p.sigma*self.pixel_size)**2 + k1) / p.intensity + k2 * (p.sigma*self.pixel_size)**4 * p.bkgstd**2 / p.intensity**2)
                uncertainty.append(p.uncertainty)
                bkgstd.append(p.bkgstd)

        self.parameter['uncertainty [nm]'] = np.asarray(uncertainty)
        self.parameter['intensity [counts]'] = np.asarray(intensity)
        self.parameter['offset [counts]'] = np.asarray(offset)
        self.parameter['x [nm]'] = np.asarray(x) * self.pixel_size
        self.parameter['y [nm]'] = np.asarray(y) * self.pixel_size
        self.parameter['sigma [nm]'] = np.asarray(sigma)
        self.parameter['bkgstd [counts]'] = np.asarray(bkgstd)
        self.parameter['frame'] = np.asarray(frame).astype(np.float32)

        for key in self.parameter:
            self.histogram_counts[key], self.histogram_bins[key] = np.histogram(self.parameter[key], bins=ParticleData.HISTOGRAM_BINS)
            self.histogram_counts[key] = self.histogram_counts[key].astype(np.float32)
            self.histogram_counts[key] = np.delete(self.histogram_counts[key], 0)
            self.histogram_bins[key] = (self.histogram_bins[key][1], self.histogram_bins[key][-1])

        self.x_min = np.min(self.parameter['x [nm]'])
        self.y_min = np.min(self.parameter['y [nm]'])
        self.x_max = np.max(self.parameter['x [nm]'])
        self.y_max = np.max(self.parameter['y [nm]'])

    def apply_filter(self, parameter_key, min_val, max_val, logic_not = False):
        vals = self.parameter[parameter_key]
        if logic_not:
            for p, v in zip(self.particles, vals):
                if min_val < v < max_val:
                    p.visible = False
        else:
            for p, v in zip(self.particles, vals):
                if v < min_val or v > max_val:
                    p.visible = False

    def set_reconstruction_roi(self, roi):
        """
        :param roi: The ROI used in the initial particle position estimation (e.g. ParticleDetectionNode). It can be saved in ParticleData in order to facilitate overlaying the final reconstruction with the corresponding region of the widefield image.
        :return:
        """
        print("Setting reconstruction ROI to: ", roi[0], roi[1], roi[2], roi[3]) ## check! todo
        self.reconstruction_roi = roi

    @staticmethod
    def from_csv(path):
        """
        :param path: path to a .csv file containing super-resolution reconstruction particle data, e.g. from ThunderStorm.
        :return: a ParticleData object.
        """
        df = pd.read_csv(path)
        particles = list()
        _idx_x = df.columns.get_loc("x [nm]")
        _idx_y = df.columns.get_loc("y [nm]")
        _idx_u = df.columns.get_loc("uncertainty [nm]")
        _idx_i = df.columns.get_loc("intensity [counts]")
        _idx_o = df.columns.get_loc("offset [counts]")
        _idx_s = df.columns.get_loc("sigma [nm]")
        _idx_b = df.columns.get_loc("bkgstd [counts]")
        _idx_f = df.columns.get_loc("frame")
        for i in range(df.shape[0]):
            _data = df.iloc[i]
            particles.append(Particle(
                _data[_idx_f],
                _data[_idx_x],
                _data[_idx_y],
                _data[_idx_s],
                _data[_idx_i],
                _data[_idx_o],
                _data[_idx_b],
                _data[_idx_u],
            ))
        particle_data_obj = ParticleData()
        particle_data_obj.pixel_size = 1
        particle_data_obj.particles = particles
        particle_data_obj.empty = False
        particle_data_obj.bake()
        print("uncertainty")
        print(particle_data_obj.parameter["uncertainty [nm]"])
        print("sigma")
        print(particle_data_obj.parameter["sigma [nm]"])
        particle_data_obj.set_reconstruction_roi([particle_data_obj.x_min, particle_data_obj.y_min, particle_data_obj.x_max, particle_data_obj.y_max])

        return particle_data_obj

    def save_as_csv(self, path):
        pd.DataFrame.from_dict(self.parameter).to_csv(path, index=False)


class Particle:
    def __init__(self, frame, x, y, sigma, intensity, offset=0, bkgstd=-1, uncertainty=-1, colour_idx=1.0):
        self.frame = frame
        self.x = x
        self.y = y
        self.sigma = sigma
        self.intensity = intensity
        self.offset = offset
        self.bkgstd = bkgstd
        self.uncertainty = uncertainty
        self.colour_idx = colour_idx
        self.visible = True

    def __str__(self):
        return f"f={self.frame}, x={self.x}, y={self.y}, sigma={self.sigma}, i={self.intensity}, offset={self.offset}, bkgstd={self.bkgstd}"



from scNodes.core.node import *


def create():
    return GetImageNode()


class GetImageNode(Node):
    title = "Dataset to image"
    group = "Converters"
    colour = (143 / 255, 143 / 255, 143 / 255, 1.0)
    sortid = 2001
    IMAGE_MODES = ["By frame", "Time projection"]
    PROJECTIONS = ["Average", "Minimum", "Maximum", "St. dev."]

    def __init__(self):
        super().__init__()
        self.size = 200
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT,parent=self)
        self.image_out = ConnectableAttribute(ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.OUTPUT, parent=self)

        self.mode = 0
        self.projection = 0
        self.frame = 0
        self.image = None
        self.load_data_source = None
        self.pixel_size = 1

        self.roi = [0, 0, 0, 0]

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.image_out.render_start()
            self.dataset_in.render_end()
            self.image_out.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()
            imgui.push_item_width(140)
            _c, self.mode = imgui.combo("Mode", self.mode, GetImageNode.IMAGE_MODES)
            self.any_change = self.any_change or _c
            imgui.pop_item_width()
            imgui.push_item_width(80)
            if self.mode == 0:
                _c, self.frame = imgui.input_int("Frame nr.", self.frame, 0, 0)
                self.any_change = self.any_change or _c
            elif self.mode == 1:
                _c, self.projection = imgui.combo("Projection", self.projection, GetImageNode.PROJECTIONS)
                if _c:
                    self.image = None
                self.any_change = self.any_change or _c
            imgui.pop_item_width()
            if self.any_change:
                self.configure_settings()

            super().render_end()

    def configure_settings(self):
        datasource = self.dataset_in.get_incoming_node()
        if datasource:
            try:
                if cfg.profiling:
                    self.profiler_count += 1
                if self.mode == 0:
                    image_in = datasource.get_image(self.frame)
                    self.image = image_in.load()
                    self.pixel_size = image_in.pixel_size
                elif self.mode == 1:
                    load_data_node = Node.get_source_load_data_node(self)
                    self.pixel_size = load_data_node.dataset.pixel_size
                    load_data_node.load_on_the_fly = False
                    self.load_data_source = load_data_node
            except Exception as e:
                cfg.set_error(e, "GetImageNode error upon attempting to gen img.\n"+str(e))
        else:
            cfg.set_error(Exception(), "GetImageNode missing input dataset.")
        self.any_change = True

    def on_update(self):
        if self.mode == 1 and self.image is None:
            if self.load_data_source is not None:
                if self.load_data_source.done_loading:
                    self.generate_projection()

    def generate_projection(self):
        data_source = self.dataset_in.get_incoming_node()
        frame = data_source.get_image(0)
        n_frames = Node.get_source_load_data_node(self).dataset.n_frames
        projection_image = np.zeros((frame.width, frame.height, n_frames))
        for i in range(n_frames):
            projection_image[:, :, i] = data_source.get_image(i).load()
        if self.projection == 0:
            self.image = np.average(projection_image, axis = 2)
        elif self.projection == 1:
            self.image = np.min(projection_image, axis = 2)
        elif self.projection == 2:
            self.image = np.max(projection_image, axis = 2)
        elif self.projection == 3:
            self.image = np.std(projection_image, axis = 2)
        self.any_change = True

    def get_image_impl(self, idx=None):
        if cfg.profiling:
            self.profiler_count -= 1
        if self.any_change:
            self.configure_settings()
        if self.image is not None:
            out_frame = Frame("virtual_frame")
            out_frame.data = self.image
            out_frame.pixel_size = self.pixel_size
            return out_frame



from scNodes.core.node import *
from skimage.feature import peak_local_max

def create():
    return ParticleDetectionNode()


class ParticleDetectionNode(Node):
    title = "Particle detection"
    group = "PSF-fitting reconstruction"
    colour = (230 / 255, 174 / 255, 13 / 255, 1.0)
    sortid = 1000

    METHODS = ["Local maximum"]
    THRESHOLD_OPTIONS = ["Value", "St. Dev.", "Mean", "Max", "Min"]

    def __init__(self):
        super().__init__()
        self.size = 290

        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent=self)
        self.localizations_out = ConnectableAttribute(ConnectableAttribute.TYPE_COORDINATES, ConnectableAttribute.OUTPUT, parent=self)


        self.method = 0
        self.thresholding = 1
        self.threshold = 100
        self.sigmas = 2.0
        self.means = 3.0
        self.n_max = 2500
        self.d_min = 1
        self.max_fac = 0.75
        self.min_fac = 5.0
        self.roi = [0, 0, 0, 0]

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.localizations_out.render_start()
            self.dataset_in.render_end()
            self.localizations_out.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            _c, self.use_roi = imgui.checkbox("use ROI", self.use_roi)
            self.any_change = self.any_change or _c
            imgui.push_item_width(160)
            _c, self.method = imgui.combo("Detection method", self.method, ParticleDetectionNode.METHODS)
            self.any_change = self.any_change or _c
            _c, self.thresholding = imgui.combo("Threshold method", self.thresholding, ParticleDetectionNode.THRESHOLD_OPTIONS)
            self.any_change = self.any_change or _c
            if self.thresholding == 0:
                _c, self.threshold = imgui.input_int("Value", self.threshold, 0, 0)
                self.any_change = self.any_change or _c
            elif self.thresholding == 1:
                _c, self.sigmas = imgui.slider_float("x Sigma", self.sigmas, 1.0, 5.0, format = "%.2f")
                self.any_change = self.any_change or _c
            elif self.thresholding == 2:
                _c, self.means = imgui.slider_float("x Mean", self.means, 0.1, 10.0, format = "%.2f")
                self.any_change = self.any_change or _c
            elif self.thresholding == 3:
                _c, self.max_fac = imgui.slider_float("x Max", self.max_fac, 0.0, 1.0, format="%.2f")
                self.any_change = self.any_change or _c
            elif self.thresholding == 4:
                _c, self.min_fac = imgui.slider_float("x Min", self.min_fac, 1.0, 10.0, format="%.2f")
                self.any_change = self.any_change or _c
            Node.tooltip("The final threshold value is determined by this factor x the metric (sigma, mean, etc.)\n"
                         "chosen above. In case of Threshold method 'Value', the value entered here is the final \n"
                         "threshold value. Note that the metric is calculated for the ROI only.")
            _c, self.n_max = imgui.slider_int("Max particles", self.n_max, 100, 2500)
            Node.tooltip("Maximum amount of particles output per frame. Ctrl + click to override the slider\n"
                         "and enter a custom value.")
            self.any_change = self.any_change or _c

            imgui.pop_item_width()
            imgui.push_item_width(100)
            _c, self.d_min = imgui.input_int("Minimum distance (px)", self.d_min, 1, 10)
            self.any_change = self.any_change or _c
            self.d_min = max([1, self.d_min])
            imgui.pop_item_width()
            super().render_end()

    def get_image_impl(self, idx=None):
        source = self.dataset_in.get_incoming_node()
        if cfg.profiling and self.FRAME_REQUESTED_BY_IMAGE_VIEWER:
            self.profiler_count += 1
        if source:
            # Find threshold value
            image_obj = source.get_image(idx)
            image = image_obj.load()
            if self.use_roi:
                image = image[self.roi[1]:self.roi[3], self.roi[0]:self.roi[2]]
            threshold = self.threshold
            if self.thresholding == 1:
                threshold = self.sigmas * np.std(image)
            elif self.thresholding == 2:
                threshold = self.means * np.mean(image)
            elif self.thresholding == 3:
                threshold = self.max_fac * np.amax(image)
            elif self.thresholding == 4:
                threshold = self.min_fac * np.abs(np.amin(image))
            # Perform requested detection method
            coordinates = peak_local_max(image, threshold_abs = threshold, num_peaks = self.n_max, min_distance = self.d_min)
            if self.use_roi:
                coordinates += np.asarray([self.roi[1], self.roi[0]])
                Node.get_source_load_data_node(self).dataset.reconstruction_roi = self.roi
            else:
                Node.get_source_load_data_node(self).dataset.reconstruction_roi = [0, 0, image.shape[0], image.shape[1]] ## todo check
            image_obj.maxima = coordinates

            return image_obj

    def get_coordinates(self, idx=None):
        try:
            if cfg.profiling:
                time_start = time.time()
                self.profiler_count += 1
            retval = self.get_image_impl(idx).maxima
            if cfg.profiling:
                self.profiler_time += time.time() - time_start
            return retval
        except Exception as e:
            cfg.set_error(e, "Error returning coordinates "+str(e))



from scNodes.core.node import *


def create():
    return BinImageNode()


class BinImageNode(Node):
    title = "Bin image"
    group = "Image processing"
    colour = (143 / 255, 143 / 255, 143 / 255, 1.0)
    sortid = 110
    MODES = ["Average", "Median", "Min", "Max", "Sum"]

    def __init__(self):
        super().__init__()
        self.size = 170
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.INPUT, self, [ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.TYPE_DATASET])
        self.output = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.OUTPUT, self, [ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.TYPE_DATASET])
        self.factor = 2
        self.mode = 0

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.output.render_start()
            self.dataset_in.render_end()
            self.output.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            imgui.push_item_width(100)
            _c, self.mode = imgui.combo("Method", self.mode, BinImageNode.MODES)
            self.any_change = _c or self.any_change
            imgui.pop_item_width()
            imgui.push_item_width(60)
            _c, self.factor = imgui.input_int("Bin factor", self.factor, 0, 0)
            self.any_change = _c or self.any_change
            self.factor = max([self.factor, 1])
            imgui.pop_item_width()
            super().render_end()

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()
        if data_source:
            image_in = data_source.get_image(idx)
            pxd = image_in.load()
            width, height = pxd.shape
            pxd = pxd[:self.factor * (width // self.factor), :self.factor * (height // self.factor)]
            if self.mode == 0:
                pxd = pxd.reshape((width // self.factor, self.factor, height // self.factor, self.factor)).mean(3).mean(1)
            elif self.mode == 1:
                pxd = pxd.reshape((width // self.factor, self.factor, height // self.factor, self.factor))
                pxd = np.median(pxd, axis=(3, 1))
            elif self.mode == 2:
                pxd = pxd.reshape((width // self.factor, self.factor, height // self.factor, self.factor)).min(3).min(1)
            elif self.mode == 3:
                pxd = pxd.reshape((width // self.factor, self.factor, height // self.factor, self.factor)).max(3).max(1)
            elif self.mode == 4:
                pxd = pxd.reshape((width // self.factor, self.factor, height // self.factor, self.factor)).sum(3).sum(1)
            return image_in




import glfw
from OpenGL.GL import *


class Window:
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        self.title = title
        self.clear_color = (0.0, 0.0, 0.0, 1.0)
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, OpenGL.GL.GL_TRUE)
        glfw.window_hint(glfw.RESIZABLE, glfw.FALSE)  ## in order to avoid an issue with pyimgui multiple contexts, window sizes are fixed.
        glfw.window_hint(glfw.SAMPLES, 4)
        self.glfw_window = glfw.create_window(self.width, self.height, self.title, None, None)
        self.focused = True
        glfw.make_context_current(self.glfw_window)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_BLEND)
        glEnable(GL_MULTISAMPLE)
        if not self.glfw_window:
            glfw.terminate()
            raise Exception("Could not initialize GLFW window.")

        # input vars
        self.cursor_pos = [0, 0]
        self.cursor_pos_previous_frame = [0, 0]
        self.cursor_delta = [0, 0]
        self.scroll_delta = [0, 0]
        self.mouse_event = MouseButtonEvent(None, None, None)
        self.key_event = KeyEvent(None, None, None)
        self.mouse_press_duration = 0.0

        # Change flags
        self.window_size_changed = False
        self.window_gained_focus = False
        self.window_gained_focus_buffer = False
        self.reset_event_timer = False

        # Aux
        self.glfw_time = 0.0
        self.time = 0.0
        self.delta_time = 0.0
        self.force_alpha_zero = False
        self.dropped_files = list()

        # default callbacks
        self.set_window_callbacks()

    def set_callbacks(self):
        # note: in current version of pyimgui, callbacks are _either_ to imgui or to app. setting callbacks takes input away from imgui.
        glfw.set_key_callback(self.glfw_window, self.key_callback)
        # glfw.set_char_callback ... not necessary atm
        glfw.set_mouse_button_callback(self.glfw_window, self.mouse_button_callback)
        glfw.set_scroll_callback(self.glfw_window, self.scroll_callback)
        glfw.set_drop_callback(self.glfw_window, self.drop_file_callback)

    def set_mouse_callbacks(self):
        glfw.set_mouse_button_callback(self.glfw_window, self.mouse_button_callback)
        glfw.set_scroll_callback(self.glfw_window, self.scroll_callback)

    def set_window_callbacks(self):
        glfw.set_window_size_callback(self.glfw_window, self.size_changed_callback)
        glfw.set_window_focus_callback(self.glfw_window, self.window_focus_callback)

    def set_key_callbacks(self):
        glfw.set_key_callback(self.glfw_window, self.key_callback)

    def on_update(self):
        current_time = glfw.get_time()
        self.delta_time = current_time - self.time
        self.time = current_time
        self.window_gained_focus = self.window_gained_focus_buffer
        self.window_gained_focus_buffer = False
        self.scroll_delta = [0.0, 0.0]
        self.dropped_files = list()
        if self.force_alpha_zero:
            self.clear_color = (*self.clear_color[0:3], 0.0)
        glClearColor(*self.clear_color)
        glClear(GL_COLOR_BUFFER_BIT)
        self.cursor_pos_previous_frame = self.cursor_pos
        self.cursor_pos = list(glfw.get_cursor_pos(self.glfw_window))
        self.cursor_delta = [-self.cursor_pos_previous_frame[0] + self.cursor_pos[0], -self.cursor_pos_previous_frame[1] + self.cursor_pos[1]]
        if self.reset_event_timer:
            self.mouse_press_duration = 0
            self.reset_event_timer = False
        if self.get_mouse_button(glfw.MOUSE_BUTTON_LEFT) or self.get_mouse_button(glfw.MOUSE_BUTTON_RIGHT) or self.get_mouse_button(glfw.MOUSE_BUTTON_MIDDLE):
            self.mouse_press_duration += self.delta_time
        else:
            self.reset_event_timer = True
        if self.focused:
            glfw.poll_events()

    def make_current(self):
        glfw.make_context_current(self.glfw_window)

    def bring_to_front(self):
        glfw.focus_window(self.glfw_window)

    def set_full_viewport(self):
        glViewport(0, 0, self.width, self.height)

    def end_frame(self):
        glfw.swap_buffers(self.glfw_window)

    def size_changed_callback(self, window, width, height):
        self.width = max([width, 1])
        self.height = max([height, 1])
        glViewport(0, 0, self.width, self.height)
        self.window_size_changed = True

    def scroll_callback(self, _, dx, dy):
        self.scroll_delta = [dx, dy]

    def mouse_button_callback(self, _, button, action, mods):
        if action == glfw.PRESS:
            self.mouse_press_duration = 0.0
        self.mouse_event = MouseButtonEvent(button, action, mods)

    def key_callback(self, window, button, scancode, action, mods):
        self.key_event = KeyEvent(button, action, mods)

    def window_focus_callback(self, window, focused):
        if focused:
            self.focused = True
            self.window_gained_focus_buffer = True
        else:
            self.focused = False

    def drop_file_callback(self, window, paths):
        self.dropped_files = paths

    def get_mouse_button(self, button):
        return glfw.get_mouse_button(self.glfw_window, button)

    def get_key_event(self, key, action, mods=0, pop_event=True):
        if self.key_event:
            if self.key_event.check(key, action, mods):
                if pop_event:
                    self.key_event = None
                return True
        return False

    def get_key(self, key):
        return glfw.get_key(self.glfw_window, key)

    def get_mouse_event(self, button, action, mods=0, max_duration=None, pop_event=True):
        if self.mouse_event:
            if self.mouse_event.check(button, action, mods):
                if pop_event:
                    self.mouse_event = None
                if max_duration is not None:  # note: event is popped regardless of max_duration
                    return self.mouse_press_duration < max_duration
                return True
        return False

    def pop_any_mouse_event(self):
        self.mouse_event = None

    def set_force_alpha_zero(self, force=True):
        self.force_alpha_zero = force


class KeyEvent:
    def __init__(self, key, action, mods):
        self.key = key
        self.action = action
        self.mods = mods

    def check(self, requested_key, requested_action, requested_mods=0):
        if self.key == requested_key:
            if self.action == requested_action:
                if self.mods == requested_mods:
                    return True
        return False

    def __str__(self):
        return f"key {self.key} action {self.action} with modifiers {self.mods}"


class MouseButtonEvent:
    def __init__(self, button, action, mods):
        self.button = button
        self.action = action
        self.mods = mods

    def check(self, requested_button, requested_action, requested_mods):
        if self.button == requested_button:
            if self.action == requested_action:
                if self.mods == requested_mods:
                    return True
        return False

    def __str__(self):
        return f"button {self.button}, action {self.action}, mods {self.mods}"
